﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Diagnostics;
using System.IO;
using System.Threading;
using System.Globalization;

namespace iikoCardISSApp
{
    public static class Extensions
    {
        public static string GetDescription(this Enum enumObj)
        {
            FieldInfo fieldInfo =
                enumObj.GetType().GetField(enumObj.ToString());

            object[] attribArray = fieldInfo.GetCustomAttributes(typeof(EnumDescription), false);
            if (attribArray.Length == 0)
                return String.Empty;
            else
            {
                EnumDescription attrib = attribArray[0] as EnumDescription;

                return attrib.Description;
            }

        }

        /// <summary>
        /// Коррекция десятичного разделителя в строке
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public static string FixSeparator(this string s)
        {
            return s.
                            Replace(".", CultureInfo.CurrentCulture.NumberFormat.NumberDecimalSeparator).
                            Replace(",", CultureInfo.CurrentCulture.NumberFormat.NumberDecimalSeparator);
        }
    }

    /// <summary>
    /// Атрибут текстового описания членов перечислений
    /// </summary>
    public class EnumDescription : Attribute
    {
        /// <summary>
        /// Конструкстор
        /// </summary>
        /// <param name="description"></param>
        public EnumDescription(string description)
        {
            m_strDescription = description;
        }

        /// <summary>
        /// Текстовое описание члена перечисления
        /// </summary>
        public string Description
        {
            get { return m_strDescription; }
        }
        private string m_strDescription;
    }

    /// <summary>
    /// Ядро службы
    /// </summary>
    public class iikoCardCore : IStartable
    {
        /// <summary>
        /// Директория запуска приложения
        /// </summary>
        static public string AppDirectory
        {
            get
            {
                return Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
            }
        }

        public iikoCardCore()
        {
            CoreState = CoreStates.Stoped;
            Queue = new Queue<AbstractOperation>();
            FullLog = new Log(Path.Combine(AppDirectory, @"logs\full.log"), this) { EveryDayLog = true };
            ErrorLog = new Log(Path.Combine(AppDirectory, @"logs\error.log"), this) { EveryDayLog = true, SendMail = true };
        }

        public enum CoreStates
        {
            [EnumDescription("Запущено")]
            Started,
            [EnumDescription("Остановлено")]
            Stoped
        }

        /// <summary>
        /// Версия ядра
        /// </summary>
        public Version Version
        {
            get { return Assembly.GetExecutingAssembly().GetName().Version; }
        }

        /// <summary>
        /// Дата файла сборки
        /// </summary>
        public DateTime VersionDate
        {
            get { return File.GetLastWriteTime(Assembly.GetExecutingAssembly().Location); }
        }

        //Полный лог
        public Log FullLog
        {
            get;
            set;
        }

        //Лог ошибок
        public Log ErrorLog
        {
            get;
            set;
        }

        //состояние
        public CoreStates CoreState
        {
            get;
            set;
        }

        /// <summary>
        /// Очередь операций службы
        /// </summary>
        public Queue<AbstractOperation> Queue
        {
            get;
            set;
        }

        /// <summary>
        /// Поток монитора очереди
        /// </summary>
        Thread queueThread;

        /// <summary>
        /// Поток получения файла конфигурации
        /// </summary>
        Thread paramThread;

        /// <summary>
        /// Конфигурационный файл
        /// </summary>
        public string ConfigFilePath
        {
            get { return configFile; }
        }
        string configFile = Path.Combine(AppDirectory, "iikoCardIS.cfg");

        /// <summary>
        /// Текущая конфигурация
        /// </summary>
        public iikoCardISConfig Config
        {
            get;
            set;
        }

        /// <summary>
        /// Основной метод мониторинга очереди операций и их выполнения
        /// </summary>
        void QueueMonitor()
        {
            while (true)
            {

                if (Queue.Count == 0)
                {
                    Thread.Sleep(10);
                    continue;
                }
                
                //берем первую операцию
                var oper = Queue.Dequeue();

                //число операций в очереди
                if (Config.MonitoringParameters.QueueCount != Queue.Count)
                {
                    lock (iikoCardISConfig.sync)
                    {
                        Config.LockLoad = true;
                        Config.MonitoringParameters.QueueCount = Queue.Count;
                        Config.Save(ConfigFilePath);
                    }
                }


                var sw = new Stopwatch();

                //и выполняем ее
                try
                {
                    sw.Start();
                    FullLog.WriteToLog(string.Format("Выполнение операции \"{0}\"...", oper.FriendlyName));

                    lock (iikoCardISConfig.sync)
                    {
                        Config.LockLoad = true;
                        Config.MonitoringParameters.CurrentOperation = oper;
                        Config.Save(ConfigFilePath);
                    }


                    oper.Execute();

                    FullLog.WriteToLog(string.Format("Операция выполнена ({0}ms)", sw.ElapsedMilliseconds));
                }
                catch (Exception ex)
                {
                    var s = string.Format("Общая ошибка операции \"{1}\" ({0}ms) - {2}", sw.ElapsedMilliseconds, oper.FriendlyName, ex.Message);

                    lock (Config)
                    {
                        Config.LockLoad = true;
                        Config.MonitoringParameters.FeedbackMessage = s;
                        Config.Save(ConfigFilePath);
                    }

                    FullLog.WriteToLog(s, ErrorLog);
                }
                finally
                {
                    FullLog.WriteToLog("---------------------------------------------------");

                    lock (Config)
                    {
                        Config.LockLoad = true;
                        Config.MonitoringParameters.CurrentOperation = null;
                        Config.Save(ConfigFilePath);
                    }
                }

            }
        }

        /// <summary>
        /// Коллекция создателей операций
        /// </summary>
        public List<AbstractOperationCreator> OperationCreators
        {
            get;
            set;
        }

        public void Start()
        {
            if (CoreState == CoreStates.Stoped)
            {
                var sw = new Stopwatch();
                sw.Start();
                FullLog.WriteToLog("*******************************************");
                FullLog.WriteToLog(String.Format("Запуск iikoCard Integration Service v{0} ({1})", Version, VersionDate.ToShortDateString()));

                if (!File.Exists(configFile))
                {
                    FullLog.WriteToLog("Файл настроек не найден", ErrorLog);
                    return;
                }
                else
                {
                    try
                    {
                        Config = (new iikoCardISConfig()).Load(configFile);
                    }
                    catch (Exception ex)
                    {
                        FullLog.WriteToLog("Ошибка первой загрузки файла настроек - "+ex.Message, ErrorLog);
                        return;
                    }
                }
                
                lock (Config)
                {
                    Config.LockLoad = true;
                    Config.MainParameters.Version = Version.ToString();
                    Config.MainParameters.VersionDate = VersionDate;
                    Config.Save(ConfigFilePath);
                }
                //поток получения настроек
                paramThread = new Thread(new ThreadStart(delegate
                    {
                        bool waserror = false;
                        while (true)
                        {
                            try
                            {
                                //файл настроек опрашивается раз в секунду для получения команд управления
                                Thread.Sleep(1000);

                                //if (Config.LockLoad) continue;

                                //lock (Config)
                                {
                                    Config = (new iikoCardISConfig()).Load(configFile);
                                    waserror = false;
                                }

                                //постановка операции в очередь из внешнего источника
                                if (Config.ControlParameters != null)
                                    if (Config.ControlParameters.PendingOperation != null)
                                    {
                                        Config.ControlParameters.PendingOperation.Core = this;
                                        Queue.Enqueue(Config.ControlParameters.PendingOperation);

                                        lock (Config)
                                        {
                                            Config.LockLoad = true;
                                            Config.ControlParameters.PendingOperation = null;
                                            Config.Save(ConfigFilePath);
                                        }
                                    }
                            }
                            catch (Exception ex)
                            {
                                if (!waserror)
                                {
                                    FullLog.WriteToLog("Ошибка чтения файла настроек - " + ex.Message, ErrorLog);
                                    waserror = true;
                                }
                            }
                        }
                    }
                    ));
                paramThread.IsBackground = true;
                paramThread.Start();
                //запуск монитора очереди
                queueThread = new Thread(new ThreadStart(QueueMonitor));
                queueThread.IsBackground = true;
                queueThread.Start();
                //запуск мониторов
                OperationCreators = new List<AbstractOperationCreator>();
                try
                {
                    OperationCreators.Add(
                        new AutoImportOperationCreator(this, new TimeMonitor() { TimeOfDay = DateTime.ParseExact(Config.MainParameters.ImportTime.Replace(":", ""), "HHmmss", null) }));
                }
                catch (Exception ex)
                {
                    FullLog.WriteToLog("!!!Автоматический импорт не включен - "+ex.Message);
                }
                try
                {
                    OperationCreators.Add(
                        new AutoExportOperationCreator(this, new TimeMonitor() { TimeOfDay = DateTime.ParseExact(Config.MainParameters.ExportTime.Replace(":", ""), "HHmmss", null) }));
                }
                catch (Exception ex)
                {
                    FullLog.WriteToLog("!!!Автоматический экспорт не включен - " + ex.Message);
                }

                CoreState = CoreStates.Started;
                FullLog.WriteToLog(String.Format("iikoCard Integration Service запущен ({0}мс)", sw.ElapsedMilliseconds));
            }
            else
            {
                FullLog.WriteToLog("!!! Запуск невозможен, т.к. iikoCard IS уже запущен");
            }
        }

        public void Stop()
        {
            throw new NotImplementedException();
        }

        public void Restart()
        {
            throw new NotImplementedException();
        }
    }
}
